#include "pwatch_com.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ncurses.h>
#include <sys/time.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>
#include <dirent.h>

#include "pwatch_proc.h"

/* Structures and enumerations */

static struct option options[] =
{
	{"refresh", required_argument, NULL, 'r'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};

enum {
	VIEW_NORMAL,
	VIEW_PROCESS
};

enum sort_type {
	UNSORTED = 0,
	ASCENDING = 1,
	DESCENDING = -1
};

struct column {
	/** The column's caption */
	const char* name;
	/** Format string for column content */
	const char* format;
	/** Width of the column, in characters. 0 = auto, -1 = hidden */
	int width;
	/** Column's sorting order */
	enum sort_type sorted;
};

static struct column columns[] =
{
	{"PID",     "%5ld", 5, UNSORTED},
	{"rbytes",  "%7s",  6, UNSORTED},
	{"wbytes",  "%7s",  6, UNSORTED},
	{"rspeed",  "%7s",  6, DESCENDING},
	{"wspeed",  "%7s",  6, UNSORTED},
	{"Process", " %s",  0, UNSORTED}
};
static const unsigned int column_count = 6;

/* Variables */
static int			max_line_len;
static struct pwatch_proc	*head;
static unsigned int		view = VIEW_NORMAL;
static int			cells = 0;
static int			selected_line = 0, last_line = 0, useful_lines = 0;
static int			max_pages = 1, page = 0;
static pid_t			selected_pid = 0;
static unsigned int		inverse = 0;


/* Functions */

static void usage(void)
{
	printf("io - Author: Catalin(ux) M BOIE\n");
	printf("Usage:\n");
	printf("	--refresh (-r) <seconds> [1]\n");
	printf("	--help (-h) - this help\n");

	printf("\n");
}

/**
 * Print a string on an entire line. The string is left justified
 * and filled with trailing spaces.
 * \param line the line number
 * \param string the string to print
 */
static void iotop_full_line(const unsigned int line, const char* string)
{
	char bg[256];
	char bg_format[8];
	unsigned int max;

	max = COLS + 1;
	if (max > sizeof(bg))
		max = sizeof(bg);

	/* Print a space character padded with COLS leading spaces */
	sprintf(bg_format, "%%%ds", COLS);
	snprintf(bg, max, bg_format, " ");
	mvaddstr(line, 0, bg);

	mvaddstr(line, 0, string);
}

static void iotop_show_help(void)
{
	char buf[256];
	int buf_len;
	int ch;
	int max;

	buf_len = sizeof(buf);

	max = max_line_len;
	if (max > buf_len)
		max = buf_len;

	/* wait for a key when calling getch */
	nodelay(stdscr, FALSE);

	erase();

	attron(COLOR_PAIR(2));

	snprintf(buf, max, "Help page - (C)opyright Catalin(ux) M. BOIE 2010\n\n");
	mvaddstr(0, 0, buf);

	/* some info */
	snprintf(buf, max,
		"Official website: http://kernel.embedromix.ro/us/\n"
		"\n");
	addstr(buf);

	/* Sorting */
	snprintf(buf, max, "Sorting:\n"); addstr(buf);
	snprintf(buf, max, "	0 - Reverse sorting		3 - by read speed\n"); addstr(buf);
	snprintf(buf, max, "	1 - by bytes read		4 - by write speed\n"); addstr(buf);
	snprintf(buf, max, "	2 - by bytes write		\n"); addstr(buf);
	snprintf(buf, max, "\n"); addstr(buf);

	/* Navigation */
	snprintf(buf, max, "Navigation:\n"); addstr(buf);
	snprintf(buf, max, "	Up/Down - go to previous/next process\n"); addstr(buf);
	snprintf(buf, max, "	PgUp/PgDown - go to previous/next page\n"); addstr(buf);
	snprintf(buf, max, "	Home/End - go to first/last page\n"); addstr(buf);
	snprintf(buf, max, "\n"); addstr(buf);

	/* Operations on selection */
	snprintf(buf, max, "Operations:\n"); addstr(buf);
	snprintf(buf, max, "	i - show file descriptor info for selected process\n"); addstr(buf);

	/* Misc */
	snprintf(buf, max, "Misc:\n"); addstr(buf);
	snprintf(buf, max, "	h - help\n"); addstr(buf);
	snprintf(buf, max, "	q - quit program\n"); addstr(buf);
	snprintf(buf, max, "	d - dump process list to file (io_<timestamp>.dump)\n"); addstr(buf);
	snprintf(buf, max, "\n"); addstr(buf);

	/* end */
	snprintf(buf, max, "Press any key to exit help.\n");
	mvaddstr(LINES - 1, 0, buf);

	refresh();
	ch = getch();

	nodelay(stdscr, TRUE);
}

/**
 * Sort by a column.
 * \param column_index the column to sort by
 */
static void iotop_sort_by_column(const unsigned int column_index)
{
	unsigned int index;

	for (index = 0; index < column_count; index++) {
		columns[index].sorted = UNSORTED;
	}

	if (inverse == 0) {
		columns[column_index].sorted = DESCENDING;
	} else {
		columns[column_index].sorted = ASCENDING;
	}
}

/**
 * Reverse sort order.
 */
static void iotop_sort_reverse(void)
{
	unsigned int index;

	for (index = 0; index < column_count; index++) {
		/* ASCENDING <--> DESCENDING */
		columns[index].sorted *= (-1);
	}
}

/**
 * Print the header for the cells view.
 */
static void iotop_show_cells_header(void)
{
	unsigned int index;
	char caption[128];
	char caption_format[8];
	int len; /* the visible length of a column caption */
	int pos;
	unsigned int width;

	/* Background */
	attron(COLOR_PAIR(3));
	iotop_full_line(1, "");

	pos = 0;
	for (index = 0; index < column_count; index++) {
		width = columns[index].width;
		switch (width) {
		case 0: /* auto */
			sprintf(caption_format, "%%s");
			len = 128;
			break;
		case -1: /* hidden */
			/* do not write anything */
			len = 0;
			break;
		default:
			sprintf(caption_format, "%%%ds", width);
			len = width;
		}
		snprintf(caption, len + 1, caption_format, columns[index].name);

		switch (columns[index].sorted) {
		case DESCENDING:
			attron(COLOR_PAIR(5));
			break;
		case ASCENDING:
			attron(COLOR_PAIR(6));
			break;
		case UNSORTED:
		default:
			attron(COLOR_PAIR(3));
		}
		mvaddstr(1, pos, caption);
		pos += len + 1;
	}
}

/*
 * Picture the entries
 */
static void iotop_show_cells(const unsigned int page)
{
	struct pwatch_proc *p;
	char buf[256];
	int buf_len;
	int line, skip, max;
	char rb[16], wb[16];
	char rs[16], ws[16];
	char format[256];
	unsigned int index;

	buf_len = sizeof(buf);

	max = max_line_len;
	if (max > buf_len)
		max = buf_len;

	/* skip is the number of entries we have to skip */
	skip = useful_lines * page;

	p = head;
	line = 0;
	while (p) {
		if (skip > 0) {
			skip--;
			p = p->next;
			continue;
		}

		pwatch_nice_output(rb, sizeof(rb), p->ts.read_bytes);
		pwatch_nice_output(wb, sizeof(wb), p->ts.write_bytes);

		pwatch_nice_output(rs, sizeof(rs), p->speed_read);
		pwatch_nice_output(ws, sizeof(ws), p->speed_write);

		strcpy(format, "");
		for (index = 0; index < column_count; index++)
			strcat(format, columns[index].format);

		snprintf(buf, max, format,
			(long int) p->pid,
			rb, wb, rs, ws, p->name);

		if (line == selected_line) {
			selected_pid = p->pid;
			attron(COLOR_PAIR(4));
		} else {
			attron(COLOR_PAIR(2));
		}
		iotop_full_line(2 + line, buf);

		if (line == last_line)
			break;

		line++;

		p = p->next;
	}
}

/*
 * Show a page of file list
 */
static void iotop_show_process(void)
{
	struct pwatch_fd *ifd;
	int lines_per_page, line, max;
	char buf[4096], flags[32];
	int buf_len;
	struct pwatch_proc *proc;

	buf_len = sizeof(buf);

	max = max_line_len;
	if (max > buf_len)
		max = buf_len;

	proc = pwatch_proc_get(head, selected_pid);
	if (!proc) {
		view = VIEW_NORMAL;
		return;
	}

	if (pwatch_proc_load_fds(proc))
		return;

	lines_per_page = LINES - 3;

	attron(COLOR_PAIR(2));

	ifd = proc->fd_head;
	line = 0;
	while (ifd) {
		if (line == lines_per_page)
			break;

		pwatch_flags(flags, sizeof(flags), ifd->flags);

		snprintf(buf, max, "%5d %16llu %16llu %13s %s",
			ifd->fd, ifd->offset, ifd->size, flags, ifd->file);
		mvaddstr(2 + line, 0, buf);

		line++;

		ifd = ifd->next;
	}
}

/*
 * Give warnings to the user if she/he does not have all things needed
 */
void iotop_warning(void)
{
	int ch, wait = 0;
	char buf[256];
	int line = 3;

	attron(COLOR_PAIR(3));

	if (pwatch_fdinfo_avail() == -1) {
		snprintf(buf, sizeof(buf),
			"Warning: Info about offset/flags of file are not"
			" available.\n"
			"         Please use a more recent kernel.");
		mvaddstr(line, 0, buf);
		line += 3;
		wait = 1;
	}

	if (pwatch_io_avail() == -1) {
		snprintf(buf, sizeof(buf),
			"Warning: Please activate TASK_IO_ACCOUNTING in kernel\n"
			"         to have information about I/O.");
		mvaddstr(line, 0, buf);
		line += 3;
		wait = 1;
	}

	if (wait == 0)
		return;

	snprintf(buf, sizeof(buf),
		"Press any key to exit this screen.");
	mvaddstr(line, 0, buf);
	line += 2;

	refresh();

	/* wait for a key when calling getch */
	nodelay(stdscr, FALSE);
	ch = getch();
	nodelay(stdscr, TRUE);
}

/*
 * Recomputes all screen stuff after some changes
 */
static void iotop_recompute_screen(void)
{
	useful_lines = LINES - 3;

	/* Compute how many pages we have */
	max_pages = (cells + useful_lines - 1) / useful_lines;

	/* Compute last_line */
	if (page < max_pages - 1)
		last_line = useful_lines - 1;
	else
		last_line = (cells - 1) % useful_lines;

	/* If list shrinked, adapt to it */
	if (page > max_pages - 1)
		page = max_pages - 1;
	if (selected_line > last_line)
		selected_line = last_line;
}

int main(int argc, char *argv[])
{
	int ch;
	unsigned int sort_by = SORT_BY_SPEED_READ;
	unsigned int re_sort = 0, force_sort;
	char status[256];
	int status_len;
	int c;
	unsigned long refresh_rate = 1;
	unsigned int refresh_count = 0, refresh_steps;
	struct timespec ts;
	unsigned int force_exit = 0;
	char error[256];
	int errcode;
	unsigned int force_redraw = 1;

	while (1) {
		int option_index = 0;

		c = getopt_long(argc, argv, "r:h", options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'r': refresh_rate = strtoul(optarg, NULL, 10); break;
			case 'h': usage(); return 0;

			case '?':
				printf("Invalid option!\n");
				usage();
				return 1;
		}
	}

	initscr();
	start_color();
	cbreak();
	noecho();

	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);

	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	init_pair(3, COLOR_WHITE, COLOR_RED);
	init_pair(4, COLOR_RED, COLOR_WHITE);
	init_pair(5, COLOR_WHITE, COLOR_BLUE);
	init_pair(6, COLOR_BLACK, COLOR_CYAN);

	iotop_warning();

	refresh_steps = refresh_rate * 20;

	force_sort = 1;

	strcpy(error, "");
	errcode = 0;

	status_len = sizeof(status);
	while (1) {
		max_line_len = COLS + 1;
		if (max_line_len > status_len)
			max_line_len = status_len;

		if ((ch = getch()) != ERR) {
			re_sort = 0;

			switch (ch) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					re_sort = 1;
					break;
				case 'd':
					pwatch_proc_dump2file(head);
					break;
				case 'h':
					iotop_show_help();
					break;
				case 'q':
				case 'Q':
					force_exit = 1;
					break;
				case KEY_NPAGE:
					if (page + 1 < max_pages) {
						page++;
						selected_line = 0;
					}
					force_redraw = 1;
					break;
				case KEY_PPAGE:
					if (page > 0) {
						page--;
						selected_line = 0;
					}
					force_redraw = 1;
					break;
				case KEY_HOME:
					page = 0;
					selected_line = 0;
					force_redraw = 1;
					break;
				case KEY_END:
					page = max_pages - 1;
					selected_line = last_line;
					force_redraw = 1;
					break;
				case KEY_UP:
					if (selected_line > 0) {
						selected_line--;
					} else if (page > 0) {
						page--;
						selected_line = useful_lines - 1;
					}
					force_redraw = 1;
					break;
				case KEY_DOWN:
					if (selected_line < last_line) {
						selected_line++;
					} else if (page < max_pages - 1) {
						page++;
						selected_line = 0;
					}
					force_redraw = 1;
					break;
				case '\r':
				case 'i':
					if (view == VIEW_NORMAL)
						view = VIEW_PROCESS;
					else
						view = VIEW_NORMAL;
					force_redraw = 1;
					break;
				case ' ':
					force_redraw = 1;
					break;
				case KEY_RESIZE:
					force_redraw = 1;
					break;
				default:
					/* do nothing */
					break;
			}

			if (force_exit == 1)
				break;
		}

		/* sorting */
		if ((force_sort == 1) || (re_sort == 1)) {
			force_sort = 0;
			re_sort = 0;

			switch (ch) {
			case '0':
				inverse = 1 - inverse;
				iotop_sort_reverse();
				break;
			case '1':
				sort_by = SORT_BY_BYTES_READ;
				iotop_sort_by_column(1);
				break;
			case '2':
				sort_by = SORT_BY_BYTES_WRITE;
				iotop_sort_by_column(2);
				break;
			case '3':
				sort_by = SORT_BY_SPEED_READ;
				iotop_sort_by_column(3);
				break;
			case '4':
				sort_by = SORT_BY_SPEED_WRITE;
				iotop_sort_by_column(4);
				break;
			}

			force_redraw = 1;
		}

		/* test if is the momment to do the refresh */
		if (refresh_count == 0) {
			if ((cells = pwatch_proc_load(&head)) < 0) {
				snprintf(error, sizeof(error), "Error getting processes info!\n");
				errcode = 1;
				break;
			}

			pwatch_proc_clean(&head);

			pwatch_proc_sort(&head, sort_by, inverse);

			force_redraw = 1;
		}

		iotop_recompute_screen();

		if (force_redraw == 1) {
			force_redraw = 0;
			erase();

			/* up status */
			attron(COLOR_PAIR(2));
			snprintf(status, max_line_len,
				"io - (C)opyright Catalin(ux aka Dino) M BOIE 20010-"
				"   [Press h for help]\n");
			mvaddstr(0, 0, status);

			/* line 2 header and body */
			attron(COLOR_PAIR(3));
			if (view == VIEW_NORMAL) {
				iotop_show_cells_header();
				iotop_show_cells(page);
			} else if (view == VIEW_PROCESS) {
				snprintf(status, max_line_len,
					"   FD           Offset             Size         Flags File");
				iotop_full_line(1, status);
				iotop_show_process();
			}

			/* down status */
			attron(COLOR_PAIR(3));
			snprintf(status, max_line_len,
				"[%lus] [%u processes] [Page %u/%u]",
				refresh_rate, cells, page + 1, max_pages);
			iotop_full_line(LINES - 1, status);

			/* Refresh screen */
			refresh();

			/* reinit counter */
			if (refresh_count == 0)
				refresh_count = refresh_steps;
		} else {
			refresh_count--;
		}

		ts.tv_sec = 0;
		ts.tv_nsec = 25000000;
		nanosleep(&ts, NULL);
	}

	endwin();

	return errcode;
}
